unit FilterNormalisedCrossCorrelation;

// NCC FORMULA
// "=  1 / (Pixels - 1) * Sum of (  ( (PixelX - AverageX) * (PixelY - AverageY) )   /   ( (StdDevX * StdDevY)  )

interface

uses
  Filter, Fparameters, Image;

type
  TFilterNormalisedCrossCorrelation = class(TFilter)
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure Run(); override;
  private
    ParameterImageToSearch, ParameterImageToFind: TParameterImage;
    ParameterImageXCentre, ParameterImageYCentre: TParameterInteger;
    OuputParameterCorrelation: TParameterSingle;
    ImageToSearch, ImageToFind: PBitmap32;
    Correlation: Single;
    procedure _run();
  end;

implementation

uses
  Math, ImageIO;

constructor TFilterNormalisedCrossCorrelation.Create;
begin
  inherited;
  ParameterImageToSearch := addParameterImage('INIMAGETOSEARCH', 'input image to search');
  ParameterImageToFind := addParameterImage('INIMAGETOFIND', 'input image to find');
  OuputParameterCorrelation := addOutputParameterSingle('CORRELATION', 'correlation between same size image 1 and 2');
  ParameterImageXCentre := addParameterInteger('X', 'the x center of image 2, in image 1', -maxint + 1, maxint, maxint);
  ParameterImageYCentre := addParameterInteger('Y', 'the y center of image 2, in image 1', -maxint + 1, maxint, maxint);
end;

destructor TFilterNormalisedCrossCorrelation.Destroy;
begin
  inherited;
end;

procedure TFilterNormalisedCrossCorrelation.Run();
begin
  ImageToSearch := ParameterImageToSearch.Image;
  ImageToFind := ParameterImageToFind.Image;
  if (ImageToSearch <> nil) and (ImageToFind <> nil) then
  begin
    _run();
  end
  else
    SetOutputParameterSingle('CORRELATION', 0);
end;

procedure TFilterNormalisedCrossCorrelation._run();
var
  N, X, Y: Integer;
  CentreX, CentreY, XOffset, YOffset: Integer;
  pSrcSearch, pSrcFind: PColor32Array;
  DataSearch, DataRef: array of Single;
  Summation, StdDevSearch, StdDevRef, AverageSearch, AverageRef: Single;
begin
  CentreX := ParameterImageXCentre.Value;
  CentreY := ParameterImageYCentre.Value;

  if (ImageToFind.Width <= ImageToSearch.Width) and (ImageToFind.Width <= ImageToSearch.Width) and
    not((CentreX > (ImageToSearch.Width - ImageToFind.Width) div 2) or (CentreX < -(ImageToSearch.Width - ImageToFind.Width) div 2) or
    (CentreY > (ImageToSearch.Height - ImageToFind.Height) div 2) or (CentreY < -(ImageToSearch.Height - ImageToFind.Height) div 2)) then
  begin
    // ******************************************************************************
    SetLength(DataSearch, ImageToFind.Width * ImageToFind.Height);
    SetLength(DataRef, ImageToFind.Width * ImageToFind.Height);

    XOffset := ((ImageToSearch.Width - ImageToFind.Width) div 2) + CentreX;
    YOffset := ((ImageToSearch.Height - ImageToFind.Height) div 2) + CentreY;

    N := 0; // byte counter

    pSrcFind := ImageToFind.Bits; // start of image to find
    for Y := 0 to ImageToFind.Height - 1 do
    begin
      pSrcSearch := ImageToSearch.Bits; // start of image to find
      Inc(pSrcSearch, ((YOffset + Y) * ImageToSearch.Width) + XOffset); // offset to line search point

      for X := 0 to ImageToFind.Width - 1 do
      begin
        DataSearch[N] := Image.Intensity(pSrcSearch^[0]); // load data to arrays
        DataRef[N] := Image.Intensity(pSrcFind^[0]);
        Inc(N); // advance pixel counter
        Inc(pSrcSearch); // advance pointers
        Inc(pSrcFind);
      end;
    end;

    // ******************************************************************************
    // calculate mean and StdDev of both arrays
    MeanAndStdDev(DataSearch, AverageSearch, StdDevSearch);
    MeanAndStdDev(DataRef, AverageRef, StdDevRef);

    // ******************************************************************************
    // calculate correlation of both arrays
    Summation := 0;
    for N := 0 to Length(DataSearch) - 1 do
    begin
      if (StdDevSearch <> 0) and (StdDevRef <> 0) then
        Summation := Summation + (((DataSearch[N] - AverageSearch) * (DataRef[N] - AverageRef)) / (StdDevSearch * StdDevRef));
    end;
    if Length(DataSearch) > 1 then
      Correlation := (1 / (Length(DataSearch) - 1)) * Summation
    else
      Correlation := 0;

    // ******************************************************************************

    // set output parameters
    SetOutputParameterSingle('CORRELATION', Correlation);

    SetLength(DataSearch, 0); // zero data arrays
    SetLength(DataRef, 0);
  end
  else
    SetOutputParameterSingle('CORRELATION', 0); // invalid image sizes
end;

end.
